package guru.kake.bst

/**
 * Implementation for a binary search tree
 * @param T Type of tree node contents. This should be a [Comparable] where [compareTo] defines the order of the tree.
 * @param list list of nodes to initialize tree from
 */
class BST<T : Comparable<T>>(list: List<T>) {
    private var root: Node<T>? = null

    init {
        val list2 = list.sorted()
        root = Node(list2[list2.size/2])
        createBST(list2, root!!)
    }

    private fun createBST(children: List<T>, parent: Node<T>) {
        val left = children.subList(0, children.size / 2)
        val right = children.subList(children.size / 2 + 1, children.size)

        if (left.isNotEmpty()) {
            parent.left = Node(left[left.size / 2])
            createBST(left, parent.left!!)
        }
        if (right.isNotEmpty()) {
            parent.right = Node(right[right.size / 2])
            createBST(right, parent.right!!)
        }
    }

    /**
     * Find [value] from tree and return it if it exist, otherwise null is returned
     * @param value
     * @return matching [T] if found or null
     */
    fun find(value: T): T? {
        var node = root
        while (true) {
            if (node!=null) {
                if (node.value < value) {
                    node = node.right
                    continue
                }
                else if (node.value > value) {
                    node = node.left
                    continue
                }
            }
            return node?.value
        }
    }

    /**
     * Debug function for finding a node
     * @param value Node to find
     * @return list of all nodes along the way to [value]
     */
    @Deprecated(message = "Will not be part of the final API")
    fun debugFind(value: T): List<T?> {
        val list = mutableListOf<T>()
        var node = root

        while (true) {
            if (node!=null) {
                println("node: ${node.value}, left: ${node.left?.value}, right: ${node.right?.value}")
                list.add(node.value)
                if (node.value < value) {
                    node = node.right
                    continue
                }
                else if (node.value > value) {
                    node = node.left
                    continue
                }
            }
            break
        }
        return list.toList()
    }

    /**
     * Node of the binary search tree
     * @param T contained type
     * @param value content of the node
     */
    class Node<T>(val value: T) {
        /** Child node which compares to less than this */
        var left: Node<T>? = null
        /** Child node which compares to more than this */
        var right: Node<T>? = null
    }
}